#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This test shows the multi-model comparison.

You will see how to:
    Perform a multi-model comparison

Author: Farid Mohammadi, M.Sc.
E-Mail: farid.mohammadi@iws.uni-stuttgart.de
Department of Hydromechanics and Modelling of Hydrosystems (LH2)
Institute for Modelling Hydraulic and Environmental Systems (IWS),
University of Stuttgart, www.iws.uni-stuttgart.de/lh2/
Pfaffenwaldring 61
70569 Stuttgart

"""

from copy import deepcopy
import numpy as np
import scipy.io as io
import pandas as pd
import joblib
import sys
sys.path.append("../../src/bayesvalidrox/")
sys.path.append("../../src/")

from bayesvalidrox.pylink.pylink import PyLinkForwardModel
from bayesvalidrox.surrogate_models.inputs import Input
from bayesvalidrox.surrogate_models.exp_designs import ExpDesigns
from bayesvalidrox.surrogate_models.surrogate_models import MetaModel
#from bayesvalidrox.surrogate_models.meta_model_engine import MetaModelEngine
from bayesvalidrox.post_processing.post_processing import PostProcessing
from bayesvalidrox.bayes_inference.bayes_inference import BayesInference
from bayesvalidrox.bayes_inference.discrepancy import Discrepancy
from bayes_inference.bayes_model_comparison import BayesModelComparison


from bayesvalidrox.surrogate_models.engine import Engine

if __name__ == "__main__":
    # Read data
    sigma = 0.6
    data = {
        'x [m]': np.linspace(0.25, 4.75, 15),
        'Z': io.loadmat('data/synth_data.mat')['d'].reshape(1, -1)[0]
        }

    # =====================================================
    # ============   COMPUTATIONAL MODELS  ================
    # =====================================================
    # Define the models options
    # ---------- Linear model  -------------
    myL2Model = PyLinkForwardModel()

    myL2Model.link_type = 'Function'
    myL2Model.py_file = 'L2_model'
    myL2Model.name = 'linear'
    myL2Model.Output.names = ['Z']
    myL2Model.observations = data

    # -- Nonlinear exponential model -------
    myNL2Model = PyLinkForwardModel()

    myNL2Model.link_type = 'Function'
    myNL2Model.py_file = 'NL2_model'
    myNL2Model.name = 'exponential'
    myNL2Model.Output.names = ['Z']
    myNL2Model.observations = data

    # ------ Nonlinear cosine model ---------
    # Data generating process
    myNL4Model = PyLinkForwardModel()

    myNL4Model.link_type = 'Function'
    myNL4Model.py_file = 'NL4_model'
    myNL4Model.name = 'cosine'
    myNL4Model.Output.names = ['Z']
    myNL4Model.observations = data

    # =====================================================
    # =========   PROBABILISTIC INPUT MODEL  ==============
    # =====================================================
    # Define model inputs
    n_sample = 10000
    # ---------- Linear model  -------------
    L2_Inputs = Input()
    L2_prior_mean = np.array([1, 0])
    L2_prior_cov = np.array(
        [[0.04, -0.007],
         [-0.007, 0.04]]
        )
    L2_input_params = np.random.multivariate_normal(
        L2_prior_mean, L2_prior_cov, size=n_sample
        )

    for i in range(L2_input_params.shape[1]):
        L2_Inputs.add_marginals()
        L2_Inputs.Marginals[i].name = f'$X_{i+1}$'
        L2_Inputs.Marginals[i].input_data = L2_input_params[:, i]

    # ------ Nonlinear exponential model ---------
    NL2_Inputs = Input()
    NL2_prior_mean = np.array([0.4, -0.3])
    NL2_prior_cov = np.array(
        [[0.003, -0.0001],
         [-0.0001, 0.03]]
        )
    NL2_input_params = np.random.multivariate_normal(
        NL2_prior_mean, NL2_prior_cov, size=n_sample
        )

    for i in range(NL2_input_params.shape[1]):
        NL2_Inputs.add_marginals()
        NL2_Inputs.Marginals[i].name = f'$X_{i+1}$'
        NL2_Inputs.Marginals[i].input_data = NL2_input_params[:, i]

    # ------ Nonlinear cosine model ---------
    NL4_Inputs = Input()
    NL4_prior_mean = np.array([2.6, 0.5, -2.8, 2.3])
    NL4_prior_cov = np.array(
        [[0.46, -0.07, 0.24, -0.14],
         [-0.07, 0.04, -0.05, 0.02],
         [0.24, -0.05, 0.30, -0.16],
         [-0.14, 0.02, -0.16, 0.30]]
        )
    NL4_input_params = np.random.multivariate_normal(
        NL4_prior_mean, NL4_prior_cov, size=n_sample
        )

    for i in range(NL4_input_params.shape[1]):
        NL4_Inputs.add_marginals()
        NL4_Inputs.Marginals[i].name = f'$X_{i+1}$'
        NL4_Inputs.Marginals[i].input_data = NL4_input_params[:, i]

    # =====================================================
    # ======  POLYNOMIAL CHAOS EXPANSION METAMODELS  ======
    # =====================================================
    # Define meta-model options
    # ---------- Linear model  -------------
    L2_MetaModelOpts = MetaModel(L2_Inputs)

    # Select if you want to preserve the spatial/temporal depencencies
    # L2_MetaModelOpts.dim_red_method = 'PCA'
    # MetaModelOpts.var_pca_threshold = 99.999
    # MetaModelOpts.n_pca_components = 10 #3, 10

    # Select your metamodel method
    L2_MetaModelOpts.meta_model_type = 'aPCE'

    # ------------------------------------------------
    # ------------- PCE Specification ----------------
    # ------------------------------------------------
    # Select the sparse least-square minimization method for
    # the PCE coefficients calculation:
    # 1)OLS: Ordinary Least Square  2)BRR: Bayesian Ridge Regression
    # 3)LARS: Least angle regression  4)ARD: Bayesian ARD Regression
    # 5)FastARD: Fast Bayesian ARD Regression
    # 6)VBL: Variational Bayesian Learning
    # 7)EBL: Emperical Bayesian Learning
    L2_MetaModelOpts.pce_reg_method = 'FastARD'

    # Specify the max degree to be compared by the adaptive algorithm:
    # The degree with the lowest Leave-One-Out cross-validation (LOO)
    # error (or the highest score=1-LOO)estimator is chosen as the final
    # metamodel. pce_deg accepts degree as a scalar or a range.
    L2_MetaModelOpts.pce_deg = np.arange(1, 12)

    # q-quasi-norm 0<q<1 (default=1)
    L2_MetaModelOpts.pce_q_norm = 1.0

    # ------------------------------------------------
    # ------ Experimental Design Configuration -------
    # ------------------------------------------------
    #L2_MetaModelOpts.add_ExpDesign()
    L2_ExpDesign = ExpDesigns(L2_Inputs)

    # One-shot (normal) or Sequential Adaptive (sequential) Design
    L2_ExpDesign.n_init_samples = 100

    # Sampling methods
    # 1) random 2) latin_hypercube 3) sobol 4) halton 5) hammersley
    # 6) chebyshev(FT) 7) grid(FT) 8)user
    L2_ExpDesign.sampling_method = 'latin_hypercube'

    # ------ Nonlinear cosine model ---------
    NL2_ExpDesign = ExpDesigns(NL2_Inputs)
    NL2_ExpDesign.method = 'normal'
    NL2_ExpDesign.n_init_samples = 100
    NL2_ExpDesign.sampling_method = 'latin_hypercube'
    NL2_MetaModelOpts = deepcopy(L2_MetaModelOpts)
    NL2_MetaModelOpts.input_obj = NL2_Inputs

    # ------ Nonlinear cosine model ---------
    NL4_MetaModelOpts = deepcopy(L2_MetaModelOpts)
    NL4_MetaModelOpts.input_obj = NL4_Inputs
    NL4_ExpDesign = ExpDesigns(NL4_Inputs)
    NL4_ExpDesign.method = 'normal'
    NL4_ExpDesign.n_init_samples = 100
    NL4_ExpDesign.sampling_method = 'latin_hypercube'

    # >>>>>> Train the Surrogates <<<<<<<<<<<
    L2_engine = Engine(L2_MetaModelOpts, myL2Model, L2_ExpDesign)
    L2_engine.start_engine()
    L2_engine.train_normal()
    NL2_engine = Engine(NL2_MetaModelOpts, myNL2Model, NL2_ExpDesign)
    NL2_engine.start_engine()
    NL2_engine.train_normal()
    NL4_engine = Engine(NL4_MetaModelOpts, myNL4Model, NL4_ExpDesign)
    NL4_engine.start_engine()
    NL4_engine.train_normal()
    #L2_MetaModel = L2_MetaModelOpts.create_metamodel(myL2Model)
    #NL2_MetaModel = NL2_MetaModelOpts.create_metamodel(myNL2Model)
    #NL4_MetaModel = NL4_MetaModelOpts.create_metamodel(myNL4Model)

    # =====================================================
    # =========  POST PROCESSING OF METAMODELS  ===========
    # =====================================================
    # ---------- Linear model  -------------
    L2_PostPCE = PostProcessing(L2_engine, name=myL2Model.name)

    # Plot to check validation visually.
    L2_PostPCE.valid_metamodel(n_samples=3)

    # Plot moments
    L2_PostPCE.plot_moments()

    # Compute and print RMSE error
    L2_PostPCE.check_accuracy(n_samples=3000)

    # Plot the sobol indices
    total_sobol = L2_PostPCE.sobol_indices()

    # ---------- Linear model  -------------
    NL2_PostPCE = PostProcessing(NL2_engine, name=myNL2Model.name)

    # Plot to check validation visually.
    NL2_PostPCE.valid_metamodel(n_samples=3)

    # Plot moments
    NL2_PostPCE.plot_moments()

    # Compute and print RMSE error
    NL2_PostPCE.check_accuracy(n_samples=3000)

    # Plot the sobol indices
    total_sobol = NL2_PostPCE.sobol_indices()

    # ------ Nonlinear cosine model ---------
    NL4_PostPCE = PostProcessing(NL4_engine, name=myNL4Model.name)

    # Plot to check validation visually.
    NL4_PostPCE.valid_metamodel(n_samples=3)

    # Plot moments
    NL4_PostPCE.plot_moments()

    # Compute and print RMSE error
    NL4_PostPCE.check_accuracy(n_samples=3000)

    # Plot the sobol indices
    total_sobol = NL4_PostPCE.sobol_indices()

    # =====================================================
    # =========  BAYESIAN MULTIMODEL COMPARISON ===========
    # =====================================================
    # ----- Define the discrepancy model -------
    sigma = np.ones(15) * np.array(sigma).flatten()
    DiscrepancyOpts = Discrepancy('')
    DiscrepancyOpts.type = 'Gaussian'
    DiscrepancyOpts.parameters = pd.DataFrame(sigma**2, columns=['Z'])

    # ----- Define the options model -------
    meta_models = {
        "linear": L2_engine,
        "exponential": NL2_engine,
        "cosine": NL4_engine
        }

    # BME Bootstrap options
    opts_bootstrap = {
        "bootstrap": True,
        "n_samples": 100,#0,#0, # TODO: difference between this and the n_bootstrap set below?
        "Discrepancy": DiscrepancyOpts,
        "emulator": True,
        "plot_post_pred": False
        }

    # Run model comparison
    BayesOpts = BayesModelComparison(
        justifiability=True,
        n_bootstrap=100,#0,#00,
        just_n_meas=2
        )
    output_dict = BayesOpts.create_model_comparison(
        meta_models,
        opts_bootstrap
        )

    # Save the results
    with open('model_comparison_output_dict.pkl', 'wb') as output:
        joblib.dump(output_dict, output, 2)
